"use strict";

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.JsDocGenerator = undefined;
exports.generate = generate;

var _typescript;

function _load_typescript() {
    return _typescript = _interopRequireWildcard(require("typescript"));
}

var _path = _interopRequireWildcard(require("path"));

var _JsDocRenderer;

function _load_JsDocRenderer() {
    return _JsDocRenderer = require("./JsDocRenderer");
}

var _util;

function _load_util() {
    return _util = require("./util");
}

var _doctrine;

function _load_doctrine() {
    return _doctrine = require("doctrine");
}

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i]; return arr2; } else { return Array.from(arr); } }

const vm = require("vm");
function generate(basePath, config, moduleName, main, options) {
    const compilerOptions = config.options;
    const compilerHost = (_typescript || _load_typescript()).createCompilerHost(compilerOptions);
    const program = (_typescript || _load_typescript()).createProgram(config.fileNames, compilerOptions, compilerHost);
    (0, (_util || _load_util()).checkErrors)((_typescript || _load_typescript()).getPreEmitDiagnostics(program));
    const compilerOutDir = compilerOptions.outDir;
    if (compilerOutDir == null) {
        throw new Error("outDir is not specified in the compilerOptions");
    }
    const generator = new JsDocGenerator(program, _path.relative(basePath, compilerOutDir), moduleName, main, program.getCommonSourceDirectory(), options, compilerOptions.baseUrl);
    for (const sourceFile of program.getSourceFiles()) {
        if (!sourceFile.isDeclarationFile) {
            generator.generate(sourceFile);
        }
    }
    return generator;
}
class JsDocGenerator {
    constructor(program, relativeOutDir, moduleName, mainFile, commonSourceDirectory, options, baseUrl) {
        this.program = program;
        this.relativeOutDir = relativeOutDir;
        this.moduleName = moduleName;
        this.mainFile = mainFile;
        this.commonSourceDirectory = commonSourceDirectory;
        this.options = options;
        this.baseUrl = baseUrl;
        this.moduleNameToResult = new Map();
        this.renderer = new (_JsDocRenderer || _load_JsDocRenderer()).JsDocRenderer(this);
        this.mainMappings = new Map();
    }
    sourceFileToModuleId(sourceFile) {
        if (sourceFile.isDeclarationFile) {
            if (sourceFile.fileName.endsWith("node.d.ts")) {
                return { id: "node", isMain: false };
            }
            let fileNameWithoutExt = sourceFile.fileName.slice(0, sourceFile.fileName.length - ".d.ts".length).replace(/\\/g, "/");
            if (this.baseUrl != null && fileNameWithoutExt.startsWith(this.baseUrl)) {
                fileNameWithoutExt = fileNameWithoutExt.substring(this.baseUrl.length + 1);
            }
            return { id: fileNameWithoutExt, isMain: false };
        }
        let sourceModuleId;
        const fileNameWithoutExt = sourceFile.fileName.slice(0, sourceFile.fileName.lastIndexOf(".")).replace(/\\/g, "/");
        const name = _path.relative(this.commonSourceDirectory, fileNameWithoutExt);
        if (this.moduleName != null) {
            sourceModuleId = this.moduleName;
            if (name !== "index") {
                sourceModuleId += "/" + this.relativeOutDir;
            }
        } else {
            sourceModuleId = this.relativeOutDir;
        }
        if (name !== "index") {
            sourceModuleId += "/" + name;
        }
        const isMain = this.mainFile == null ? fileNameWithoutExt.endsWith("/main") : `${fileNameWithoutExt}.js`.indexOf(_path.posix.relative(this.relativeOutDir, this.mainFile)) !== -1;
        if (isMain) {
            sourceModuleId = this.moduleName;
        }
        return { id: sourceModuleId, isMain };
    }
    generate(sourceFile) {
        if (sourceFile.text.length === 0) {
            return;
        }
        const moduleId = this.sourceFileToModuleId(sourceFile);
        this.currentSourceModuleId = moduleId.id;
        const classes = [];
        const functions = [];
        const members = [];
        (0, (_util || _load_util()).processTree)(sourceFile, node => {
            if (node.kind === (_typescript || _load_typescript()).SyntaxKind.InterfaceDeclaration || node.kind === (_typescript || _load_typescript()).SyntaxKind.ClassDeclaration) {
                const descriptor = this.processClassOrInterface(node);
                if (descriptor != null) {
                    classes.push(descriptor);
                }
            } else if (node.kind === (_typescript || _load_typescript()).SyntaxKind.FunctionDeclaration) {
                const descriptor = this.describeFunction(node);
                if (descriptor != null) {
                    functions.push(descriptor);
                }
            } else if (moduleId.isMain && node.kind === (_typescript || _load_typescript()).SyntaxKind.ExportDeclaration) {
                this.handleExportFromMain(node);
                return true;
            } else if (node.kind === (_typescript || _load_typescript()).SyntaxKind.SourceFile) {
                return false;
            } else if (node.kind === (_typescript || _load_typescript()).SyntaxKind.VariableStatement) {
                const descriptor = this.describeVariable(node);
                if (descriptor != null) {
                    members.push(descriptor);
                }
            } else if (node.kind === (_typescript || _load_typescript()).SyntaxKind.EnumDeclaration) {
                const descriptor = this.describeEnum(node);
                if (descriptor != null) {
                    members.push(descriptor);
                }
            }
            return true;
        });
        const existingPsi = this.moduleNameToResult.get(moduleId.id);
        if (existingPsi == null) {
            this.moduleNameToResult.set(moduleId.id, { classes, functions, members });
        } else {
            var _existingPsi$classes, _existingPsi$function, _existingPsi$members;

            (_existingPsi$classes = existingPsi.classes).push.apply(_existingPsi$classes, classes);
            (_existingPsi$function = existingPsi.functions).push.apply(_existingPsi$function, functions);
            (_existingPsi$members = existingPsi.members).push.apply(_existingPsi$members, members);
        }
    }
    handleExportFromMain(node) {
        const moduleSpecifier = node.moduleSpecifier;
        const exportClause = node.exportClause;
        if (exportClause == null || moduleSpecifier == null) {
            return;
        }
        if (moduleSpecifier.kind !== (_typescript || _load_typescript()).SyntaxKind.StringLiteral) {
            return;
        }
        const filePath = moduleSpecifier.text;
        if (!filePath.startsWith(".")) {
            return;
        }
        const fullFilename = _path.posix.resolve(_path.posix.dirname(node.getSourceFile().fileName), filePath) + ".ts";
        const sourceFile = this.program.getSourceFile(fullFilename);
        if (sourceFile == null) {
            return;
        }
        const names = [];
        for (const e of exportClause.elements) {
            if (e.kind === (_typescript || _load_typescript()).SyntaxKind.ExportSpecifier) {
                names.push(e.name.text);
            } else {
                console.error(`Unsupported export element: ${e.getText(e.getSourceFile())}`);
            }
        }
        this.mainMappings.set(this.sourceFileToModuleId(sourceFile).id, names);
    }
    getTypeNamePathByNode(node) {
        if (node.kind === (_typescript || _load_typescript()).SyntaxKind.UnionType) {
            return this.typesToList(node.types, node);
        } else if (node.kind === (_typescript || _load_typescript()).SyntaxKind.FunctionType) {
            return ["callback"];
        } else if (node.kind === (_typescript || _load_typescript()).SyntaxKind.NumberKeyword) {
            return ["number"];
        } else if (node.kind === (_typescript || _load_typescript()).SyntaxKind.StringKeyword) {
            return ["string"];
        } else if (node.kind === (_typescript || _load_typescript()).SyntaxKind.BooleanKeyword) {
            return ["boolean"];
        } else if (node.kind === (_typescript || _load_typescript()).SyntaxKind.NullKeyword) {
            return ["null"];
        } else if (node.kind === (_typescript || _load_typescript()).SyntaxKind.UndefinedKeyword) {
            return ["undefined"];
        } else if (node.kind === (_typescript || _load_typescript()).SyntaxKind.LiteralType) {
            const text = node.literal.text;
            return [`"${text}"`];
        } else if (node.kind === (_typescript || _load_typescript()).SyntaxKind.TypeLiteral) {
            // todo
            return ['Object.<string, any>'];
        }
        const type = this.program.getTypeChecker().getTypeAtLocation(node);
        return type == null ? null : this.getTypeNames(type, node);
    }
    typesToList(types, node) {
        const typeNames = [];
        for (const type of types) {
            if (type.kind == null) {
                const name = this.getTypeNamePath(type);
                if (name == null) {
                    throw new Error(`Cannot get name for ${node.getText(node.getSourceFile())}`);
                }
                typeNames.push(name);
            } else {
                const name = this.getTypeNamePathByNode(type);
                if (name == null) {
                    throw new Error(`Cannot get name for ${node.getText(node.getSourceFile())}`);
                }
                typeNames.push.apply(typeNames, _toConsumableArray(name));
            }
        }
        return typeNames;
    }
    getTypeNames(type, node) {
        if (type.flags & (_typescript || _load_typescript()).TypeFlags.UnionOrIntersection && !(type.flags & (_typescript || _load_typescript()).TypeFlags.Enum) && !(type.flags & (_typescript || _load_typescript()).TypeFlags.EnumLiteral) && !(type.flags & (_typescript || _load_typescript()).TypeFlags.Boolean) && !(type.flags & (_typescript || _load_typescript()).TypeFlags.BooleanLiteral)) {
            return this.typesToList(type.types, node);
        }
        let result = this.getTypeNamePath(type);
        if (result == null) {
            throw new Error("Cannot infer getTypeNamePath");
        }
        const typeArguments = type.typeArguments;
        if (typeArguments != null) {
            const subTypes = [];
            for (const type of typeArguments) {
                subTypes.push.apply(subTypes, _toConsumableArray(this.getTypeNames(type, node)));
            }
            return [{ name: result, subTypes: subTypes }];
        }
        return [result];
    }
    getTypeNamePath(type) {
        if (type.flags & (_typescript || _load_typescript()).TypeFlags.Boolean) {
            return "boolean";
        }
        if (type.flags & (_typescript || _load_typescript()).TypeFlags.Void) {
            return "void";
        }
        if (type.flags & (_typescript || _load_typescript()).TypeFlags.Null) {
            return "null";
        }
        if (type.flags & (_typescript || _load_typescript()).TypeFlags.String) {
            return "string";
        }
        if (type.flags & (_typescript || _load_typescript()).TypeFlags.Number) {
            return "number";
        }
        if (type.flags & (_typescript || _load_typescript()).TypeFlags.Undefined) {
            return "undefined";
        }
        if (type.flags & (_typescript || _load_typescript()).TypeFlags.Any) {
            return "any";
        }
        if (type.flags & (_typescript || _load_typescript()).TypeFlags.Literal) {
            return `"${type.value}"`;
        }
        const symbol = type.symbol;
        if (symbol == null || symbol.declarations == null || symbol.declarations.length === 0) {
            return null;
        }
        const valueDeclaration = symbol.valueDeclaration || (symbol.declarations == null || symbol.declarations.length === 0 ? null : symbol.declarations[0]);
        if ((_typescript || _load_typescript()).getCombinedModifierFlags(valueDeclaration) & (_typescript || _load_typescript()).ModifierFlags.Ambient) {
            // Error from lib.es5.d.ts
            return symbol.name;
        }
        let typeSourceParent = valueDeclaration;
        while (typeSourceParent != null) {
            if (typeSourceParent.kind === (_typescript || _load_typescript()).SyntaxKind.ModuleDeclaration && (typeSourceParent.flags & (_typescript || _load_typescript()).NodeFlags.NestedNamespace) <= 0) {
                const m = typeSourceParent;
                const sourceModuleId = m.name.text;
                if (typeSourceParent.flags & (_typescript || _load_typescript()).NodeFlags.Namespace) {
                    return `${sourceModuleId}:${symbol.name}`;
                } else {
                    return `module:${sourceModuleId}.${symbol.name}`;
                }
            } else if (typeSourceParent.kind === (_typescript || _load_typescript()).SyntaxKind.SourceFile) {
                const sourceModuleId = this.sourceFileToModuleId(typeSourceParent).id;
                return `module:${sourceModuleId}.${symbol.name}`;
            }
            typeSourceParent = typeSourceParent.parent;
        }
        console.warn(`Cannot find parent for ${symbol}`);
        return null;
    }
    describeEnum(node) {
        const flags = (_typescript || _load_typescript()).getCombinedModifierFlags(node);
        if (!(flags & (_typescript || _load_typescript()).ModifierFlags.Export)) {
            return null;
        }
        const type = {
            names: ["number"]
        };
        const name = node.name.text;
        const moduleId = this.computeTypePath();
        const id = `${moduleId}.${name}`;
        const properties = [];
        for (const member of node.members) {
            const name = member.name.text;
            properties.push({
                name: name,
                kind: "member",
                scope: "static",
                memberof: id,
                type: type
            });
        }
        // we don't set readonly because it is clear that enum is not mutable
        // e.g. jsdoc2md wil add useless "Read only: true"
        return {
            node: node,
            id: id,
            name: name,
            longname: id,
            kind: "enum",
            scope: "static",
            memberof: moduleId,
            type: type,
            properties: properties
        };
    }
    describeVariable(node) {
        const flags = (_typescript || _load_typescript()).getCombinedModifierFlags(node);
        if (!(flags & (_typescript || _load_typescript()).ModifierFlags.Export)) {
            return null;
        }
        const declarations = node.declarationList == null ? null : node.declarationList.declarations;
        if (declarations == null || declarations.length !== 1) {
            return null;
        }
        const declaration = declarations[0];
        if (declaration.type == null) {
            return null;
        }
        const existingJsDoc = (_JsDocRenderer || _load_JsDocRenderer()).JsDocRenderer.getComment(node);
        const jsDoc = existingJsDoc == null ? null : (0, (_doctrine || _load_doctrine()).parse)(existingJsDoc, { unwrap: true });
        if (JsDocGenerator.isHidden(jsDoc)) {
            return null;
        }
        let types;
        const type = this.program.getTypeChecker().getTypeAtLocation(declaration);
        if (type.symbol != null && type.symbol.valueDeclaration != null) {
            types = [this.getTypeNamePath(type)];
        } else {
            types = this.getTypeNamePathByNode(declaration.type);
        }
        // NodeFlags.Const on VariableDeclarationList, not on VariableDeclaration
        return { types, node, name: declaration.name.text, isConst: (node.declarationList.flags & (_typescript || _load_typescript()).NodeFlags.Const) > 0 };
    }
    //noinspection JSMethodCanBeStatic
    describeFunction(node) {
        const flags = (_typescript || _load_typescript()).getCombinedModifierFlags(node);
        if (!(flags & (_typescript || _load_typescript()).ModifierFlags.Export)) {
            return null;
        }
        const existingJsDoc = (_JsDocRenderer || _load_JsDocRenderer()).JsDocRenderer.getComment(node);
        const jsDoc = existingJsDoc == null ? null : (0, (_doctrine || _load_doctrine()).parse)(existingJsDoc, { unwrap: true });
        return JsDocGenerator.isHidden(jsDoc) ? null : { name: node.name.text, node: node, tags: [], jsDoc };
    }
    static isHidden(jsDoc) {
        if (jsDoc == null) {
            return false;
        }
        for (const tag of jsDoc.tags) {
            if (tag.title === "internal" || tag.title === "private") {
                return true;
            }
        }
        return false;
    }
    processClassOrInterface(node) {
        const flags = (_typescript || _load_typescript()).getCombinedModifierFlags(node);
        if (!(flags & (_typescript || _load_typescript()).ModifierFlags.Export)) {
            return null;
        }
        const nodeDeclaration = node;
        const existingJsDoc = (_JsDocRenderer || _load_JsDocRenderer()).JsDocRenderer.getComment(node);
        const jsDoc = existingJsDoc == null ? null : (0, (_doctrine || _load_doctrine()).parse)(existingJsDoc, { unwrap: true });
        if (JsDocGenerator.isHidden(jsDoc)) {
            return null;
        }
        const className = nodeDeclaration.name.text;
        const clazz = node;
        let parents = [];
        if (clazz.heritageClauses != null) {
            for (const heritageClause of clazz.heritageClauses) {
                if (heritageClause.types != null) {
                    for (const type of heritageClause.types) {
                        const typeNamePath = this.getTypeNamePathByNode(type);
                        if (typeNamePath != null) {
                            parents = parents.concat(typeNamePath);
                        }
                    }
                }
            }
        }
        const methods = [];
        const properties = [];
        for (const member of nodeDeclaration.members) {
            if (member.kind === (_typescript || _load_typescript()).SyntaxKind.PropertySignature) {
                const p = this.describeProperty(member, node.kind === (_typescript || _load_typescript()).SyntaxKind.ClassDeclaration);
                if (p != null) {
                    properties.push(p);
                }
            } else if (member.kind === (_typescript || _load_typescript()).SyntaxKind.PropertyDeclaration) {
                const p = this.describeProperty(member, node.kind === (_typescript || _load_typescript()).SyntaxKind.ClassDeclaration);
                if (p != null) {
                    properties.push(p);
                }
            } else if (member.kind === (_typescript || _load_typescript()).SyntaxKind.GetAccessor) {
                const p = this.describeProperty(member, node.kind === (_typescript || _load_typescript()).SyntaxKind.ClassDeclaration);
                if (p != null) {
                    properties.push(p);
                }
            } else if (member.kind === (_typescript || _load_typescript()).SyntaxKind.MethodDeclaration || member.kind === (_typescript || _load_typescript()).SyntaxKind.MethodSignature) {
                const m = this.renderMethod(member);
                if (m != null) {
                    methods.push(m);
                }
            }
        }
        methods.sort((a, b) => {
            let weightA = a.isProtected ? 100 : 0;
            let weightB = b.isProtected ? 100 : 0;
            // do not reorder getFeedURL/setFeedURL
            weightA += trimMutatorPrefix(a.name).localeCompare(trimMutatorPrefix(b.name));
            return weightA - weightB;
        });
        return {
            modulePath: this.computeTypePath(),
            name: className,
            node, methods, properties, parents,
            isInterface: node.kind === (_typescript || _load_typescript()).SyntaxKind.InterfaceDeclaration
        };
    }
    describeProperty(node, isParentClass) {
        const flags = (_typescript || _load_typescript()).getCombinedModifierFlags(node);
        if (flags & (_typescript || _load_typescript()).ModifierFlags.Private) {
            return null;
        }
        if (this.options.access === "public" && flags & (_typescript || _load_typescript()).ModifierFlags.Protected) {
            return null;
        }
        const name = node.name.text;
        let types;
        if (node.type == null) {
            const type = this.program.getTypeChecker().getTypeAtLocation(node);
            if (type == null) {
                return null;
            }
            types = this.getTypeNames(type, node);
        } else {
            types = this.getTypeNamePathByNode(node.type);
        }
        let isOptional = node.questionToken != null;
        let defaultValue = null;
        const initializer = node.initializer;
        if (initializer != null) {
            if (initializer.expression != null || initializer.kind === (_typescript || _load_typescript()).SyntaxKind.NoSubstitutionTemplateLiteral) {
                defaultValue = initializer.getText();
            } else {
                try {
                    const sandbox = { sandboxVar: null };
                    vm.runInNewContext(`sandboxVar=${initializer.getText()}`, sandbox);
                    const val = sandbox.sandboxVar;
                    if (val === null || typeof val === "string" || typeof val === "number" || "boolean" || Object.prototype.toString.call(val) === "[object Array]") {
                        defaultValue = val;
                    } else if (val) {
                        console.warn(`unknown initializer for property ${name}: ${val}`);
                    }
                } catch (e) {
                    console.info(`exception evaluating initializer for property ${name}`);
                    defaultValue = initializer.getText();
                }
            }
        }
        isOptional = isOptional || defaultValue != null || types.indexOf("null") !== -1;
        if (!isOptional && isParentClass && (flags & (_typescript || _load_typescript()).ModifierFlags.Readonly) > 0) {
            isOptional = true;
        }
        return { name, types, node, isOptional, defaultValue };
    }
    renderMethod(node) {
        // node.flags doesn't report correctly for private methods
        const flags = (_typescript || _load_typescript()).getCombinedModifierFlags(node);
        if (flags & (_typescript || _load_typescript()).ModifierFlags.Private) {
            return null;
        }
        if (this.options.access === "public" && flags & (_typescript || _load_typescript()).ModifierFlags.Protected) {
            return null;
        }
        const tags = [];
        const isProtected = (flags & (_typescript || _load_typescript()).ModifierFlags.Protected) > 0;
        if (isProtected) {
            tags.push(`@protected`);
        }
        const name = node.name.text;
        const existingJsDoc = (_JsDocRenderer || _load_JsDocRenderer()).JsDocRenderer.getComment(node);
        const jsDoc = existingJsDoc == null ? null : (0, (_doctrine || _load_doctrine()).parse)(existingJsDoc, { unwrap: true });
        return JsDocGenerator.isHidden(jsDoc) ? null : { name, tags, isProtected, node, jsDoc };
    }
    computeTypePath() {
        return `module:${this.currentSourceModuleId}`;
    }
}
exports.JsDocGenerator = JsDocGenerator;
function trimMutatorPrefix(name) {
    if (name.length > 4 && name[3] === name[3].toUpperCase() && (name.startsWith("get") || name.startsWith("set"))) {
        return name[3].toLowerCase() + name.substring(4);
    }
    return name;
}
//# sourceMappingURL=JsDocGenerator.js.map